# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

__version__ = "0.1.6"
__authors__ = "Sergio Fernández"
__license__ = "GNU Library or Lesser General Public License (LGPL) v3"
__contact__ = "sergio@wikier.org"
__date__    = "2010-04-21"
__agent__   = "djubby %s (http://djubby.googlecode.com/)" % __version__

# Some common stuff for network:
import socket
import urllib2
socket.setdefaulttimeout(10)

# Making imports easier for users:
from dispatcher import dispatcher
from configuration import Configuration




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

import logging
from django.http import HttpResponse, HttpResponseRedirect, Http404
from configuration import Configuration
from resource import Resource
from http import Http303, get_preferred_prefix, get_preferred_output, get_mimetype, url_handler
from urllib2 import URLError

def dispatcher(request, ref=None):
    logging.debug("Dispatching request on '%s'..." % ref)
    conf = Configuration()
    if (ref == None or len(ref) == 0):
        index = conf.get_value("indexResource")
        index = index.replace(conf.get_value("datasetBase"), conf.get_value("webBase"))
        logging.debug("Redirecting to the index resource...")
        return HttpResponseRedirect(index)
    else:
        try:
            uri, prefix = url_handler(ref)
            resource = Resource(uri)
        except ValueError, ve:
            logging.error("Error processing request for '%s': %s" % (ref, str(ve)))
            raise Http404(ve)
        except URLError, ue:
            logging.error("Error retrieving data for '%s': %s" % (ref, str(ue)))
            raise Http404(ue)

        if (prefix == None):
            prefix = get_preferred_prefix(request)
            get_url = getattr(resource, "get_%s_url" % prefix)
            url = get_url()
            logging.debug("Redirecting to the %s representation of %s: %s" % (prefix, uri, url))
            return Http303(url)
        else:         
            output = get_preferred_output(request, prefix)
            func = getattr(resource, "get_%s_%s" % (prefix, output))
            mimetype = get_mimetype(prefix, output)            
            logging.debug("Returning the %s representation of %s serialized as %s" % (prefix, uri, output))       
            return HttpResponse(func(), mimetype=mimetype)




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

import os
import logging
from django.conf import settings
from rdflib.Graph import ConjunctiveGraph
import rdf
import ns

class Configuration:
    """Configuration using the Borg design pattern"""

    __shared_state = { "data" : None, "path" : None, "graph" : None, "endpoint" : None}

    def __init__(self, path=None):
        self.__dict__ = self.__shared_state
        if (self.data == None):
            if (path == None):
                raise ValueError("djubby's configuration MUST be initialized a first time, read http://code.google.com/p/djubby/wiki/GettingStarted")
            else:
                self.path = os.path.abspath(path)
                logging.debug("Reading djubby's configuration from %s..." % self.path)
                if (not os.path.exists(self.path)):
                    raise ValueError("Not found a proper file at '%s' with a configuration for djubby. Please, provide a right path" % self.path)

                data = ConjunctiveGraph()
                data.bind("conf", ns.config) 
                try:
                    data.load(path, format='n3') 
                except Exception, e:
                    raise ValueError("Not found a proper N3 file at '%s' with a configuration for djubby. Please, provide a valid N3 file" % self.path)

                self.data = data
                try:
                    self.graph = self.get_value("sparqlDefaultGraph")
                    self.endpoint = self.get_value("sparqlEndpoint")
                except Exception, e:
                    raise ValueError("Not found the graph not the endpoint that it's supposed djubby have to query. Please, provide a right donfiguration")

                logging.info("Using <%s> as default graph to query the endpoint <%s>" % (self.graph, self.endpoint))
                self.__class__.__dict__['_Configuration__shared_state']["data"] = data #FIXME

    def get_values(self, prop):
        return rdf.get_values(self.data, predicate=ns.config[prop])

    def get_value(self, prop):
        return rdf.get_value(self.data, predicate=ns.config[prop])




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

from rdflib import Namespace

rdf    = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
rdfs   = Namespace("http://www.w3.org/2000/01/rdf-schema#")
xsd    = Namespace("http://www.w3.org/2001/XMLSchema#")
owl    = Namespace("http://www.w3.org/2002/07/owl#")
config = Namespace("http://richard.cyganiak.de/2007/pubby/config.rdf#")
dc     = Namespace("http://purl.org/dc/elements/1.1/")
dct    = Namespace("http://purl.org/dc/terms/")
skos   = Namespace("http://www.w3.org/2004/02/skos/core#")
foaf   = Namespace("http://xmlns.com/foaf/0.1/")




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

from configuration import Configuration
import mimeparse
from sets import Set
from django.http import HttpResponseRedirect
from django.utils.datastructures import MultiValueDictKeyError

formats = {
            "data" : { "default":"application/rdf+xml", "xml":"application/rdf+xml", "n3":"text/n3" },
            "page" : { "default":"text/html", "html":"text/html", "xhtml":"application/xhtml+xml" }

          }

def get_supported_prefixes():
    return formats.keys()

def get_supported_formats():
    fs = Set()
    for f in formats.itervalues():
        for m in f.itervalues():
            fs.add(m)
    return list(fs)

def get_supported_outputs():
    outputs = Set()
    for f in formats.itervalues():
        for o in f.iterkeys():
            if (not o == "default"):
                outputs.add(o)
    return list(outputs)

def get_mimetype(prefix, output):
    if (not prefix in formats):
        prefix = "data"
    if (not output in formats[prefix]):
        output = "default"
    return formats[prefix][output]

def get_prefix(mimetype):
    for prefix, f in formats.items():
        mimes = f.values()
        if mimetype in mimes:
            return prefix
    return "data"

def get_preferred_format(request):
    try:
        accept = request.META["HTTP_ACCEPT"]
    except KeyError:
        accept = formats["data"]["xml"]
    return mimeparse.best_match(get_supported_formats(), accept)

def get_preferred_prefix(request):
    return get_prefix(get_preferred_format(request))

def get_preferred_output(request, prefix):
    if (prefix == "page"):
        return "html"
    else:
        try:
            output = request.GET["output"]
            if (output in get_supported_outputs()):
                return output
            else:
                return "xml"
        except MultiValueDictKeyError:
            format = get_preferred_format(request)
            for output, mimetype in formats["data"].items():
                if (output != "default"):
                    if (mimetype == format):
                        return output
            return "xml"

def get_document_url(uri, prefix, conf=None):
    if (conf == None):
        conf = Configuration()
    return uri.replace(conf.get_value("datasetBase"), conf.get_value("webBase")).replace(conf.get_value("webResourcePrefix"), "%s/" % prefix)

def url_handler(ref):
    uri = None
    prefix = None
    conf = Configuration()
    datasetBase = conf.get_value("datasetBase")
    webBase = conf.get_value("webBase")
    resourcePrefix = conf.get_value("webResourcePrefix")

    if (ref.startswith(resourcePrefix)):
        prefix = None
        uri = datasetBase + ref
        return uri, prefix
    else:
        splitted = ref.split("/")
        prefix = splitted[0]
        if (prefix in get_supported_prefixes()):
            uri = datasetBase + ref.replace(prefix+"/", conf.get_value("webResourcePrefix"))
            return uri, prefix
        else:
            raise ValueError("Unsupportet type '%s'" % splitted[0])

class Http303(HttpResponseRedirect):
    status_code = 303




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

from rdflib import URIRef

class URI:

    def __init__(self, uri, label=None):
        if (type(uri)==URIRef):
            self.uri = unicode(uri)
        else:
            self.uri = uri
        self.label = label

    def __str__(self):
        return self.uri

    def __cmp__(self, o):
        return cmp(self.uri, o.uri)

    def __eq__(self, o):
        return self.uri.__eq__(o.uri)

    def __hash__(self):
        return self.uri.__hash__()

def get_values(graph, subject=None, predicate=None):
    return graph.objects(subject=str2uri(subject), predicate=predicate)

def get_value(graph, subject=None, predicate=None, lang=None):
    if (lang == None):
        try:
            return str(get_values(graph, subject, predicate).next())
        except StopIteration:
            return ""
    else:
        values = get_values(graph, subject, predicate)
        for value in values:
            if (value.language == lang):
                return value
        return ""

def get_predicates(graph, subject=None):
    return graph.predicate_objects(str2uri(subject))

def str2uri(uri):
    if (type(uri)==str or type(uri)==unicode):
        return URIRef(uri)
    else:
        return uri

def uri2curie(uri, namespaces):
    url, fragment = splitUri(uri)
    for prefix, ns in namespaces: #improve the performace of this operation       
        if (unicode(ns) == url):
            return "%s:%s" % (prefix, fragment)
    return uri

def splitUri(uri):
    if ("#" in uri):
        splitted = uri.split("#")
        return ("%s#"%splitted[0], splitted[1])
    else:
        splitted = uri.split("/")
        return ("/".join(splitted[:-1])+"/", splitted[-1])




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.

import os
import logging
from configuration import Configuration
from SPARQLWrapper import SPARQLWrapper, JSON
from django.template import Template, Context
import rdf
import ns
from rdflib import URIRef
from rdflib import Literal
from http import get_document_url

class Resource:

    queries = {
                "ask"      : "ASK { GRAPH <%s> { <%s> ?p ?o } }",
                "describe" : "DESCRIBE <%s> FROM <%s>"
              }

    def __init__(self, uri):
        logging.debug("Trying to build resource with URI <%s>..." % uri)
        self.uri = uri
        self.conf = Configuration()
        self.graph = self.conf.graph
        self.endpoint = self.conf.endpoint
        if (not self.__ask__()):
            raise ValueError("Resource wirh URI <%s> not found on this dataset" % self.uri)
        else:
            logging.info("Successfully found the resource with URI <%s> on this dataset" % self.uri)

    def get_triples(self):
        sparql = SPARQLWrapper(self.endpoint)
        sparql.setQuery(self.queries["describe"] % (self.uri, self.graph))
        g = sparql.query().convert()
        logging.debug("Returning %d triples describing resource <%s>" % (len(g), self.uri))
        #FIXME: enrich with metadata
        for prefix, namespace in self.conf.data.namespaces():
            g.bind(prefix, namespace) 
        return g

    def get_uri(self):
        return uri

    def get_data_url(self):
        return get_document_url(self.uri, "data")

    def get_page_url(self):
        return get_document_url(self.uri, "page")

    def get_data(self):
        return get_data_xml()

    def get_data_xml(self):
        g = self.get_triples()
        return g.serialize(format="pretty-xml")

    def get_data_n3(self):
        g = self.get_triples()
        return g.serialize(format="n3")

    def get_page(self):
        return get_page_html()

    def get_page_html(self):
        g = self.get_triples()
        tpl = Template(self.__read_template__())

        data = {}
        data["uri"] = self.uri
        lang = self.conf.get_value("defaultLanguage")
        data["lang"] = lang
        label = rdf.get_value(g, self.uri, ns.rdfs["label"], lang)
        if (len(label)>0):
            data["label"] = label
        else:
            data["label"] = self.uri
        datasetBase = self.conf.get_value("datasetBase")
        webBase = self.conf.get_value("webBase")
        data["data"]  = self.get_data_url()
        data["project"] = self.conf.get_value("projectName")
        data["homelink"] = self.conf.get_value("projectHomepage")
        data["endpoint"] = self.conf.get_value("sparqlEndpoint")
        depiction = rdf.get_value(g, self.uri, ns.foaf["depiction"])
        if (len(depiction)>0):
            data["depiction"] = depiction

        data["rows"] = self.__get_rows__(g)

        ctx = Context(data)
        return tpl.render(ctx)

    def __ask__(self):
        sparql = SPARQLWrapper(self.endpoint)
        query = self.queries["ask"] % (self.graph, self.uri)
        sparql.setQuery(query)
        sparql.setReturnFormat(JSON)
        results = sparql.query().convert()
        if (results.has_key("boolean")):
            # expected answer according SPARQL Protocol
            if (not results["boolean"]):
                return False
        elif (results.has_key("results") and results["results"].has_key("bindings") and len(results["results"]["bindings"])>0):
            # I don't know why, but virtuoso sometimes uses __ask_retval
            # http://docs.openlinksw.com/virtuoso/rdfsparql.html
            if (not bool(results["results"]["bindings"][0]["__ask_retval"]["value"])):
                return False
        else:
            return False
        return True

    def __get_rows__(self, g):
        rows = {}
        for p, o in rdf.get_predicates(g, self.uri):
            prop = rdf.URI(unicode(p), rdf.uri2curie(unicode(p), self.conf.data.namespaces()))
            if (not rows.has_key(prop)):
                rows[prop] = []
            if (type(o) == URIRef):
                rows[prop].append(rdf.URI(unicode(o), rdf.uri2curie(unicode(o), self.conf.data.namespaces())))
            elif (type(o) == Literal):
                item = {}
                item["literal"] = unicode(o)
                if (o.language):
                    item["language"] = o.language
                if (o.datatype):
                    item["datatype"] = rdf.uri2curie(o.datatype, self.conf.data.namespaces())
                rows[prop].append(item)
            else:
                rows[prop].append(o)
        return rows

    def __read_template__(self, name="resource"):
        path = "%s/../tpl/%s.tpl" % (os.path.dirname(__file__), name)
        logging.debug("Reading template '%s' from %s" % (name, path))
        f = open(path, "r")
        content = f.read()
        f.close()
        return content




# -*- coding: utf-8 -*-

# This file is part of djubby <http://djubby.googlecode.com/>,
# a Linked Data frontend for SPARQL endpoints for the Django Web framework
#
# Copyright (C) 2010 Sergio Fernández
#
# Djubby is free software: you can redistribute it and/or modify it 
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Djubby is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Djubby. If not, see <http://www.gnu.org/licenses/>.


from ez_setup import use_setuptools
use_setuptools()
from setuptools import setup, find_packages
 
setup(
      name             = "djubby",
      version          =  "0.1.6",
      description      = "A Linked Data frontend for SPARQL endpoints for Django",
      license          = "GNU Library or Lesser General Public License (LGPL) v3",
      author           = "Sergio Fernández",
      author_email     = "sergio@wikier.org",
      url              = "http://djubby.googlecode.com/",
      download_url     = "http://code.google.com/p/djubby/downloads/list",
      platforms        = ["any"],
      packages         = ["djubby"],
      requires         = ["rdflib", "SPARQLWrapper", "django", "mimeparse"], 
      install_requires = ["rdflib >= 2.4.0", "SPARQLWrapper >= 1.3.2", "django >= 1.1.0", "mimeparse >= 0.1.2"],
      classifiers      = [
                            "Development Status :: 4 - Beta ",
                            "Intended Audience :: Developers",
                            "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
                            "Operating System :: OS Independent",
                            "Programming Language :: Python :: 2.6",
                         ],
      keywords         = "python django rdf sparql linkeddata",
      scripts          = ["ez_setup.py"],
      data_files       = [ ("tpl", ["tpl/resource.tpl"]) ]
)

